"""This module provides a Pythonic interface to the TeamCity RESTful API."""

# Import standard modules
from pathlib import Path
from string import Template
from typing import Any, Dict, Optional

# Import third-party modules
import requests
from requests import certs, codes, exceptions  # type: ignore[attr-defined]
from requests.auth import HTTPBasicAuth

# Import internal modules
from .lang import BatCaveError, BatCaveException, FROZEN, BATCAVE_HOME


class TeamCityError(BatCaveException):
    """TeamCity Exceptions.

    Attributes:
        BAD_CONFIG: The requested configuration was not found.
    """
    BAD_CONFIG = BatCaveError(1, Template('Invalid configuration ID: $id'))


class TCBuildConfig:
    """Class to create a universal abstract interface for a TeamCity build configuration."""

    def __init__(self, server: 'TeamCityServer', config_id: str, /):
        """
        Args:
            server: The Teamcity console containing this object.
            config_id: The configuration ID.

        Attributes:
            config_id: The value of the config_id argument.
            info: The configuration info returned from the Teamcity API.
            _server: The value of the server argument.
        """
        self._server = server
        self.config_id = config_id
        try:
            self.info = self._server.api_call('get', 'buildTypes/id:' + self.config_id)
        except exceptions.HTTPError as err:
            raise TeamCityError(TeamCityError.BAD_CONFIG, id=self.config_id) from err

    def __str__(self):
        return str(self.info)

    def __getattr__(self, attr: str):
        if attr in self.info:
            return self.info[attr]
        return self._server.api_call('get', f'buildTypes/id:{self.config_id}/{attr}')


class TeamCityServer:
    """Class to create a universal abstract interface for a TeamCity server."""

    _CA_CERT = (BATCAVE_HOME / 'cacert.pem') if FROZEN else Path(certs.where())  # type: ignore[attr-defined]

    def __init__(self, host: str, /, user: str, passwd: str, port: str = '80'):
        """
        Args:
            host: The server hosting the Teamcity instance.
            user: The Teamcity user for API access.
            password: The Teamcity password for API access.
            port (optional, default='80'): The port on which the instance is hosted.

        Attributes:
            url: The URL to the Teamcity RESTful API.
            auth: The authorization credentials for the Teamcity server.
        """
        self.url = f'http://{host}:{port}/httpAuth/app/rest/'
        self.auth = HTTPBasicAuth(user, passwd)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    build_configs = property(lambda s: s.api_call('get', 'buildTypes')['buildType'], doc='A read-only property which returns a list of the build configurations.')
    users = property(lambda s: s.api_call('get', 'users')['user'], doc='A read-only property which returns a list of the TeamCity users.')
    groups = property(lambda s: s.api_call('get', 'userGroups')['group'], doc='A read-only property which returns a list of the TeamCity groups.')

    def api_call(self, call_type: str, api_call: str, /, **params) -> Dict[str, Any]:
        """Provide an interface to the TeamCity RESTful API.

        Args:
            call_type: The API call type.
            api_call: The API call.
            **params (optional, default={}): Any parameters to pass to the API call.

        Returns:
            The result of the API call.

        Raises:
            An error unless the result of the API call is OK.
        """
        caller = getattr(requests, call_type)
        result = caller(self.url + api_call, auth=self.auth, headers={'Content-Type': 'application/json', 'Accept': 'application/json'}, json=params, verify=self._CA_CERT)
        if result.status_code != codes.ok:  # pylint: disable=no-member
            raise result.raise_for_status()
        return result.json()

    def create_group(self, name: str, key: Optional[str] = None, /, description: str = '') -> Dict[str, Any]:
        """Create a user group.

        Args:
            name: The name of user group.
            key (optional, default=None): The key name to use for the user group.
            description (optional, default=''): The description to use for the user group.

        Returns:
            The result of the API call to create the user group.
        """
        if not key:
            key = name.upper().replace(' ', '_')
        return self.api_call('post', 'userGroups', name=name, key=key, description=description)

    def create_user(self, username: str, /) -> Dict[str, Any]:
        """Create a user.

        Args:
            username: The name of user.

        Returns:
            The result of the API call to create the user group.
        """
        return self.api_call('post', 'users', username=username)

    def get_build_config(self, config: str, /) -> 'TCBuildConfig':
        """Get the named build configuration.

        Args:
            config: The build configuration to return.

        Returns:
            The requested build configuration.
        """
        return TCBuildConfig(self, config)

# cSpell:ignore cacert batcave



"""This module provides a simplified interface to the PyQt5 module.


Attributes:
    MessageType (Enum): The message types supported by the MessageBox class.
"""

# pylint: disable=too-few-public-methods,invalid-name

# Import standard modules
import sys
from enum import Enum
from pathlib import Path
from string import Template
from typing import cast, Any, Callable, List, Optional, TextIO, Tuple

# Import GUI framework and widgets
from PyQt5.QtCore import QEvent  # pylint: disable=no-name-in-module
from PyQt5.QtGui import QCloseEvent, QIcon, QImage  # pylint: disable=no-name-in-module
from PyQt5.QtWidgets import QDialog, QFileDialog, QMainWindow, QMessageBox, QWidget  # pylint: disable=no-name-in-module

# Import internal modules
from .lang import MsgStr, FROZEN, BATCAVE_HOME
from .version import AppVersion, VersionStyle

MessageType = Enum('MessageType', ('about', 'info', 'question', 'warning', 'error', 'results'))

APP_TITLE = 'The App'
ABOUT_INFO = 'No About Info Available'


class Title(MsgStr):
    """Class to extend the MsgStr class for handling MessageBox titles.

    Attributes:
        _messages: The different message types.
    """
    _messages = {'about': f'About {APP_TITLE}',
                 'info': f'{APP_TITLE} Information',
                 'question': f'{APP_TITLE} Question',
                 'warning': f'{APP_TITLE} Warning',
                 'error': f'{APP_TITLE} ERROR',
                 'results': 'Search Results'}

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs (optional, default={}): The list of arguments to pass to the base class.
        """
        super().__init__(transform='title', **kwargs)


class Brief(MsgStr):
    """Class to extend the MsgStr class for handling short window messages.

    Attributes:
        _messages: The different message types.
    """
    _messages = {'about': ABOUT_INFO,
                 'info': f'{APP_TITLE} Information',
                 'question': f'{APP_TITLE} Question',
                 'warning': f'{APP_TITLE} Warning',
                 'error': "I'm Sorry Dave, I Can't Do That",
                 'results': 'Search Results'}


class Message(MsgStr):
    """Class to extend the MsgStr class for handling messages.

    Attributes:
        _messages: The different message types.
    """
    _messages = {'INTERNAL_ERROR': Template('INTERNAL ERROR: Unknown $what Action'),
                 'MISSING_INFO': Template('You must $how a $what')}


class BatCaveValidator:
    """Class to support control validation."""

    def __init__(self, callback: Callable, false_val: Any, how: str, what: str):
        """
        Args:
            callback: The callback to use for validation.
            false_val: The value to indicate a false validation.
            how: The verb to use in the error message.
            what: The noun to use in the error message.

        Attributes:
            callback: The value of the callback argument.
            false_val: The value of the false_val argument.
            how: The value of the how argument.
            what: The value of the what argument.
        """
        self.callback = callback
        self.false_val = false_val
        self.how = how
        self.what = what


class BatCaveGUIOutput:
    """Class to manage output to a widget."""

    def __init__(self, widget: QWidget, /):
        """
        Args:
            widget: The widget to which output will be sent.

        Attributes:
            widget: The value of the widget argument.
        """
        self.widget = widget

    def write(self, output: str, /) -> None:
        """Write to the widget.

        Args:
            output: The text to write to the widget.

        Returns:
            Nothing.
        """
        self.widget.append(output)  # type: ignore[attr-defined]


class BatCaveBaseGUI:
    """The base class for the simplified GUI support.

    This class cannot be used by itself but must be paired with another PyQt class.
    """

    def __init__(self, _unused_parent: Optional[QWidget] = None, title: str = '', icon: Optional[QIcon] = None):
        """
        Args:
            _unused_parent: The parent for this object.
            title (optional, default=''): The title for this object.
            icon (optional, default=None): The icon for this object.

        Attributes:
            validators: The list of widget validators.
            _saved_output_streams: The list of saved output streams.
        """
        self.setupUi(self)  # type: ignore[attr-defined]  # pylint: disable=no-member
        self.validators: List[Any] = []
        self._saved_output_streams: Tuple = tuple()
        if title:
            self.setWindowTitle(title)  # type: ignore[attr-defined]  # pylint: disable=no-member
        if icon:
            self.setWindowIcon(QIcon(str(find_image(icon))))  # type: ignore[attr-defined,arg-type]  # type: ignore[attr-defined]  # pylint: disable=no-member

    def closeEvent(self, event: QCloseEvent, /) -> None:
        """Overload of standard Qt method called when the object is closed.

        Returns:
            Nothing.
        """
        if event.type() == QEvent.Close:
            if self._saved_output_streams:
                (sys.stdout, sys.stderr) = self._saved_output_streams  # pylint: disable=unbalanced-tuple-unpacking

    def redirect_output(self, widget: QWidget, /) -> None:
        """Redirect stdout and stderr to the specified widget.

        Returns:
            Nothing.
        """
        self._saved_output_streams = (sys.stdout, sys.stderr)
        sys.stdout = sys.stderr = cast(TextIO, BatCaveGUIOutput(widget))

    def validate(self) -> bool:
        """Run the validators for the object.

        Returns:
            True if all the validators are True, False otherwise.
        """
        for validator in self.validators:
            if validator.callback() == validator.false_val:
                MessageBox(self, Message(how=validator.how, what=validator.what).MISSING_INFO, MessageType.error).exec()
                return False
        return True


class BatCaveMainWindow(QMainWindow, BatCaveBaseGUI):
    """This class provides functionality for a main window."""

    def __init__(self, parent: Optional[QWidget] = None, title: str = '', version: Optional[AppVersion] = None, icon: Optional[QIcon] = None):
        """
        Args:
            parent (optional, default=None): The parent for the window.
            title (optional, default=''): The title for the window.
            icon (optional, default=None): The icon for the window.
        """
        super().__init__(parent, title=(title if title else version.get_info(VersionStyle.brief) if version else 'Main'), icon=icon)  # type: ignore[call-arg]
        global ABOUT_INFO, APP_TITLE  # pylint: disable=global-statement
        if version:
            APP_TITLE = version.title
            ABOUT_INFO = version.get_info(VersionStyle.about_box)
        self.actionAbout.triggered.connect(self.OnAbout)  # type: ignore[attr-defined]

    def OnAbout(self) -> None:
        """Show the about box.

        Returns:
            Nothing.
        """
        MessageBox(self, ABOUT_INFO, MessageType.about).exec()


class BatCaveDialog(QDialog, BatCaveBaseGUI):
    """This class provides functionality for a dialog box window."""
    def accept(self) -> bool:  # type: ignore[override]
        """Overload of standard Qt method called when the dialog is accepted.

        Returns:
            The result of the validate method.
        """
        return self.validate()

    def onGetDirectory(self) -> None:
        """Show a simplified directory dialog.

        Returns:
            Nothing.
        """
        edit_control = getattr(self, self.sender().objectName().replace('btn', 'edt'))
        if dirpath := Path(QFileDialog.getExistingDirectory(self)):
            edit_control.setText(dirpath)

    def onGetFile(self, file_filter: Optional[str] = None, /) -> None:
        """Show a simplified file dialog.

        Args:
            file_filter (optional, default=None): The file filter to pass to the standard Qt file dialog.

        Returns:
            Nothing.
        """
        edit_control = getattr(self, self.sender().objectName().replace('btn', 'edt'))
        if filepath := Path(QFileDialog.getOpenFileName(self, filter=str(file_filter))[0]):
            edit_control.setText(filepath)


class MessageBox(QMessageBox):
    """This class provides functionality for a simplified message box.

    Attributes:
        _MESSAGE_ICONS: The supported message box icons.
    """
    _MESSAGE_ICONS = {MessageType.about: QMessageBox.Information,
                      MessageType.info: QMessageBox.Information,
                      MessageType.question: QMessageBox.Question,
                      MessageType.warning: QMessageBox.Warning,
                      MessageType.error: QMessageBox.Critical,
                      MessageType.results: QMessageBox.Information}

    def __init__(self, parent: QWidget | BatCaveBaseGUI, message: str, /, msg_type: MessageType = MessageType.info, *, detail: str = '', image: Optional[str] = None):
        """
        Args:
            parent: The parent for the message box.
            message: The message for the message box.
            msg_type (optional, default=MessageType.info): The message type to display in the message box.
            detail (optional, default=''): The detail information to put in the message box.
            image (optional, default=None): The image to display in the message box.
        """
        super().__init__(cast(QWidget, parent))
        if image:
            self.setWindowIcon(QIcon(find_image(image)))
        self.setWindowTitle(getattr(Title(), msg_type.name))
        self.setIcon(self._MESSAGE_ICONS[msg_type])
        self.setText(getattr(Brief(), msg_type.name))
        self.setInformativeText(message)
        self.adjustSize()
        if detail:
            self.setDetailedText(detail)


def find_image(name: str, /) -> QImage:
    """Locate the image based on whether the application has been frozen.

    Args:
        name: The name of the image to locate.

    Returns:
        The image object.
    """
    image_dir = BATCAVE_HOME if FROZEN else (BATCAVE_HOME / 'img')
    return cast(QImage, [f for f in image_dir.glob(name + '.*')][0])  # pylint: disable=unnecessary-comprehension

# cSpell:ignore batcave



"""This module provides improved date/time support."""

# Import standard modules
from copy import deepcopy
from datetime import datetime, timezone
from string import Template
from typing import Optional, TypeAlias, Union

# Import third-party modules
from dateutil.parser import parse as parse_timestamp
from google.api_core.datetime_helpers import DatetimeWithNanoseconds

# Import internal modules
from .lang import BatCaveError, BatCaveException

BatCaveDateTimeType: TypeAlias = Union[DatetimeWithNanoseconds, float, str, datetime, 'BatCaveDateTime']


class TimeError(BatCaveException):
    """Class for time exceptions."""
    BAD_CONVERSION_TYPE = BatCaveError(1, Template('Unable to convert $type to TMDateTime'))


class BatCaveDateTime:
    """Class to manage date/time values."""
    _OUTPUT_FORMAT = '%Y-%m-%d-%H-%M-%S%Z'
    _INPUT_FORMAT = '%Y-%m-%d-%H-%M-%S%z'

    def __init__(self, time_info: Optional[BatCaveDateTimeType] = None):
        self._datetime = self._to_dt(deepcopy(time_info) if time_info else datetime.now(timezone.utc))

    def __eq__(self, other: object):
        if isinstance(other, BatCaveDateTime):
            return self.datetime == other.datetime
        return self.datetime == other

    def __ne__(self, other: object):
        if isinstance(other, BatCaveDateTime):
            return self.datetime != other.datetime
        return self.datetime != other

    def __gt__(self, other: object):
        if isinstance(other, BatCaveDateTime):
            return self.datetime > other.datetime
        return self.datetime > other

    def __lt__(self, other: object):
        if isinstance(other, BatCaveDateTime):
            return self.datetime < other.datetime
        return self.datetime < other

    def __str__(self):
        return self._dt_to_str(self.datetime)

    def __sub__(self, other: object):
        if isinstance(other, BatCaveDateTime):
            return self.datetime - other.datetime
        return self.datetime - other

    def _dt_to_str(self, dt_obj: 'datetime') -> str:
        return dt_obj.strftime(self._OUTPUT_FORMAT)

    def _str_to_dt(self, dt_str: str) -> datetime:
        if len(dt_str.split('-')) == 6:
            return datetime.strptime(dt_str.removesuffix('UTC') + '+0000', self._INPUT_FORMAT)
        return parse_timestamp(dt_str).astimezone(timezone.utc)

    def _to_dt(self, time_info: BatCaveDateTimeType) -> datetime:
        if isinstance(time_info, DatetimeWithNanoseconds):
            return self._str_to_dt(self._dt_to_str(time_info))
        if isinstance(time_info, float):
            return datetime.utcfromtimestamp(time_info)
        if isinstance(time_info, str):
            return self._str_to_dt(time_info)
        if isinstance(time_info, datetime):
            if not time_info.tzinfo:
                time_info = self._str_to_dt(self._dt_to_str(time_info) + 'UTC')
            return time_info
        if isinstance(time_info, BatCaveDateTime):
            return time_info.datetime
        raise TimeError(TimeError.BAD_CONVERSION_TYPE, type=type(time_info))

    datetime = property(lambda s: s._datetime, doc='A read-only property which returns the datetime representation.')

    def ctime(self) -> str:
        """Return the ctime representation."""
        return self.datetime.ctime()

# cSpell:ignore astimezone dateutil



"""Unit tests for the statemachine module."""

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name,protected-access
# flake8: noqa

from unittest import main, TestCase

from batcave.statemachine import StateMachine, StateMachineError, StateStatus


class TestStateMachine(TestCase):
    STATES = ('state1', 'state2')

    def setUp(self):
        if StateMachine._DEFAULT_STATEFILE.exists():
            StateMachine._DEFAULT_STATEFILE.unlink()
        if StateMachine._DEFAULT_LOCKFILE.exists():
            StateMachine._DEFAULT_LOCKFILE.unlink()

    def tearDown(self):
        self.setUp()

    def test_statemachine_happypath(self):
        with StateMachine(self.STATES) as sm:
            self.assertEqual(sm.status, StateStatus.exited)
            self.assertEqual(sm.state, 'None')
            self.assertTrue(StateMachine._DEFAULT_STATEFILE.exists())
            self.assertTrue(StateMachine._DEFAULT_LOCKFILE.exists())
            for state in self.STATES:
                sm.enter_next_state()
                self.assertEqual(sm.status, StateStatus.entering)
                self.assertEqual(sm.state, state)
                sm.exit_state()
                self.assertEqual(sm.status, StateStatus.exited)
                self.assertEqual(sm.state, state)
            self.assertRaises(StateMachineError, sm.enter_next_state)
            try:
                sm.enter_next_state()
            except StateMachineError as err:
                self.assertEqual(StateMachineError.DONE.code, err.code)
        self.assertTrue(StateMachine._DEFAULT_STATEFILE.exists())
        self.assertFalse(StateMachine._DEFAULT_LOCKFILE.exists())

    def test_statemachine_savestate(self):
        for _unused_state in self.STATES:
            with StateMachine(self.STATES) as sm:
                sm.enter_next_state()
                sm.exit_state()

    def test_statemachine_badentry(self):
        with StateMachine(self.STATES) as sm:
            sm.enter_next_state()
            self.assertRaises(StateMachineError, sm.enter_next_state)
            try:
                sm.enter_next_state()
            except StateMachineError as err:
                self.assertEqual(StateMachineError.BAD_ENTRY.code, err.code)

    def test_statemachine_badexit(self):
        with StateMachine(self.STATES) as sm:
            sm.enter_next_state()
            sm.exit_state()
            self.assertRaises(StateMachineError, sm.exit_state)
            try:
                sm.exit_state()
            except StateMachineError as err:
                self.assertEqual(StateMachineError.BAD_EXIT.code, err.code)


if __name__ == '__main__':
    main()



"""This module provides utilities for working with versions.

Attributes:
    PYQT_LOADED (bool/str): If not False then it is the string version of the PyQt API.
    VersionStyle (Enum): The version output styles.
"""

# Import standard modules
from dataclasses import dataclass
from enum import Enum
from sys import version as sys_version

# Import third-party modules
try:
    import PyQt5.QtCore as pyqt
except ImportError:
    PYQT_LOADED = ''
else:
    PYQT_LOADED = pyqt.__dict__['PYQT_VERSION_STR']

# Import internal modules
from .platarch import Platform

VersionStyle = Enum('VersionStyle', ('full', 'brief', 'one_line', 'about_box'))


@dataclass(frozen=True)
class AppVersion:
    """This class holds the version information for the running application.

        Attributes:
            title: The application title.
            version: The application version.
            build_date: The application build date.
            build_name: The application build name.
            copyright: The application copyright.
    """
    title: str
    version: str
    build_date: str = ''
    build_name: str = ''
    copyright: str = ''

    def get_info(self, style: VersionStyle = VersionStyle.full, /, plattype: str = 'batcave_run', extra_info: str = '') -> str:
        """Get the version information about the currently running application.

        Args:
            style (optional, default=full): The format for the version string to be returned.
            plattype (optional, default='batcave_run'): The platform type for the architecture information in the version string.
            extra_info (optional, default=''): A line to append after the version info but before the copyright.

        Returns:
            Returns the version string.
        """
        plat = getattr(Platform(), plattype)
        if style == VersionStyle.brief:
            return f'{self.title} {self.version}{extra_info} on {plat}'
        if style == VersionStyle.one_line:
            version_string = f'{self.title} {self.version}{extra_info} '
            if self.build_name:
                version_string += f'(Build: {self.build_name}) '
            if self.build_date:
                version_string += f'[{self.build_date}] '
            return f'{version_string}on {plat}'

        info = ['' if (style == VersionStyle.about_box) else f'{self.title} {self.version}']
        if self.build_name:
            info.append('Build: ' + self.build_name)
        if self.build_date:
            info.append('Date: ' + self.build_date)
        info += ['Platform: ' + plat,
                 'Python Version: ' + sys_version]
        if PYQT_LOADED:
            info.append('PyQt Version: ' + PYQT_LOADED)
        from . import cms  # pylint: disable=import-outside-toplevel
        if cms.P4_LOADED or cms.GIT_LOADED:
            info.append('CMS Support:')
            if cms.P4_LOADED:
                info.append('    P4Python: ' + cms.P4_LOADED)
            if cms.GIT_LOADED:
                info.append('    GitPython API: ' + cms.GIT_LOADED)
        if extra_info:
            info += [extra_info]
        if self.copyright:
            info.append(self.copyright)
        return '\n'.join(info)

# cSpell:ignore pyqt batcave platarch plattype



"""This module provides a Pythonic interface to work with Internet Information Server."""

# Import standard modules
from os import getenv
from pathlib import Path
from string import Template
from typing import Dict, List, Optional, Type
from xml.etree.ElementTree import fromstring as xmlparse_str, fromstringlist as xmlparse_list

# Import internal modules
from .sysutil import syscmd, CMDError
from .lang import bool_to_str, str_to_pythonval, BatCaveError, BatCaveException, CommandResult, PathName


class AppCmdError(BatCaveException):
    """appcmd Exceptions.

    Attributes:
        APPCMD_ERROR: There was an error executing appcmd.exe.
    """
    APPCMD_ERROR = BatCaveError(1, Template('Error running appcmd: $message'))


class IISConfigurationError(BatCaveException):
    """IIS Configuration Exceptions.

    Attributes:
        PARSE_ERROR: There was an error locating the specified configuration section.
    """
    PARSE_ERROR = BatCaveError(1, Template("Unable to locate '$expected' in configuration"))


class IISAdvancedLogError(BatCaveException):
    """IIS AdvancedLog Exceptions.

    Attributes:
        BAD_FIELD: There was an attempt to add a non-existent field.
        NOT_INSTALLED: Advanced logging is not installed.
    """
    BAD_FIELD = BatCaveError(1, Template('Attempt to add non-existent field: $field'))
    NOT_INSTALLED = BatCaveError(2, 'IIS Advanced Logging is not installed on the target system')


class IISObject:
    """Class to create a universal abstract interface for an IIS object."""

    def __init__(self, name: str, iis_ref: 'IISInstance', /):
        """
        Args:
            name: The name of the object.
            iis_ref: A reference to the IIS owner instance.

        Attributes:
            _iis_ref: The value of the iis_ref argument.
            _name: The value of the name argument.
        """
        self._iis_ref = iis_ref
        self._name = name

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    iis_ref = property(lambda s: s._iis_ref, doc='A read-only property which returns the reference to the IIS object.')
    name = property(lambda s: s._name, doc='A read-only property which returns the name of the IIS object.')

    def manage_item(self, action: str, /) -> None:
        """Perform action on this object.

        Args:
            action: The action to perform on this object.

        Returns:
            Nothing.
        """
        self.iis_ref.manage_item(action, type(self), self.name)

    def start(self) -> None:
        """Start this object.

        Returns:
            Nothing.
        """
        self.manage_item('start')

    def stop(self) -> None:
        """Stop this object.

        Returns:
            Nothing.
        """
        try:
            self.manage_item('stop')
        except AppCmdError as err:
            if err.vars['returncode'] != 1062:
                raise


class VirtualDirectory(IISObject):
    """Class to create a universal abstract interface for an IIS virtual directory."""


class WebApplication(IISObject):
    """Class to create a universal abstract interface for an IIS web application."""


class WebApplicationPool(IISObject):
    """Class to create a universal abstract interface for an IIS web application pool."""


class WebSite(IISObject):
    """Class to create a universal abstract interface for an IIS website."""


class IISInstance:  # pylint: disable=too-many-public-methods
    """Class to create a universal abstract interface for an IIS instance.

    Attributes:
        _IIS_TYPE_MAP: Maps IIS object types to appcmd subcommands.
    """
    _IIS_TYPE_MAP = {VirtualDirectory: 'VDIR',
                     WebApplication: 'APP',
                     WebApplicationPool: 'APPPOOL',
                     WebSite: 'SITE'}

    def __init__(self, hostname: Optional[str] = None, /, *, remote_powershell: Optional[bool] = None):
        """
        Args:
            hostname (optional, default=localhost): The name of the IIS server hosting the instance.
            remote_powershell (optional): Determines if PowerShell remoting is used when executing appcmd against a remote server.
                Defaults to False for hostname is None, otherwise defaults to True.

        Attributes:
            _hostname: The value of the hostname argument.
            _remote_powershell: The resolved value of the remote_powershell argument.
        """
        self._hostname = hostname
        self._remote_powershell = (True if (remote_powershell is None) else remote_powershell) if hostname else False

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> List[str]:
        tag = attr.upper()[0:-1]
        result = [i.attrib[f'{tag}.NAME'].rstrip('/') for i in xmlparse_list(appcmd('list', attr, hostname=self.hostname, remote_powershell=self._remote_powershell)).findall(tag)]
        if not result:
            raise AttributeError(attr)
        return result

    hostname = property(lambda s: s._hostname, doc='A read-only property which returns the hostname of the IIS server.')

    def create_virtual_dir(self, vdir_name: str, /, vdir_location: PathName, website: str) -> VirtualDirectory:
        """Create the specified virtual directory in the IIS instance.

        Args:
            vdir_name: The name of the virtual directory to create.
            vdir_location: The physical path for the virtual directory.
            website: The website for which to create the virtual directory.

        Returns:
            The newly created virtual directory.
        """
        self.manage_item('add', VirtualDirectory, f'/app.name:{website}/', f'/path:/{vdir_name}', f'/physicalPath:{vdir_location}')
        return self.get_virtual_dir(vdir_name)

    def create_webapp(self, app_name: str, appdir: PathName, website: str, pool: Optional[WebApplicationPool] = None) -> WebApplication:
        """Create the specified web application in the IIS instance.

        Args:
            app_name: The name of the web application to create.
            appdir: The physical path for the web application.
            website: The website for which to create the virtual directory.
            pool (optional, default=None): If not None, the application pool to which to assign the web application.

        Returns:
            The newly created web application.
        """
        self.manage_item('add', WebApplication, f'/site.name:{website}', f'/path:/{app_name}/', f'/physicalPath:{appdir}', f'/applicationPool:{pool}' if pool else '')
        return self.get_webapp(app_name)

    def create_webapp_pool(self, pool_name: str, /) -> WebApplicationPool:
        """Create the specified web application pool in the IIS instance.

        Args:
            pool_name: The name of the web application to create.

        Returns:
            The newly created web application pool.
        """
        self.manage_item('add', WebApplicationPool, f'/name:{pool_name}')
        return self.get_webapp_pool(pool_name)

    def exists(self) -> bool:
        """Test for existence of the IIS instance.

        Returns:
            True if the instance exists, False, otherwise.
        """
        try:
            str(self.get_configuration_section('log'))
        except CMDError as err:
            if err.code == CMDError.CMD_ERROR.code:
                return False
            raise
        return True

    def get_advanced_logger(self, path: Optional[PathName] = None, logtype: str = 'server', set_location: str = 'apphost') -> 'IISAdvancedLogger':
        """Get the advanced logger object from the IIS instance.

        Args:
            path (optional, default=None): If not None, return the advanced logger from the specified path.
            logtype (optional, default='server'): Use the specified set location to search for the advanced logger.
            set_location (optional, default='apphost'): Use the specified set location to search for the advanced logger.

        Returns:
            The specified advanced logger.
        """
        return IISAdvancedLogger(path, logtype=logtype, set_location=set_location, hostname=self.hostname)

    advanced_logger = property(get_advanced_logger, doc='A read-only property which returns the advanced logger object from the IIS instance.')

    def get_configuration_section(self, name: str, /, path: Optional[PathName] = None, set_location: str = 'apphost') -> 'IISConfigurationSection':
        """Get the named configuration section from the IIS configuration files.

        Args:
            name: The name of the configuration section to return.
            path (optional, default=None): If not None, return the configuration section from the specified path.
            set_location (optional, default='apphost'): Use the specified set location to search for the configuration section.

        Returns:
            The specified configuration section.
        """
        return IISConfigurationSection(name, path=path, set_location=set_location, hostname=self.hostname, remote_powershell=self._remote_powershell)

    def get_virtual_dir(self, vdir_name: str, /) -> VirtualDirectory:
        """Get the specified virtual directory from the IIS instance.

        Args:
            vdir_name: The name of the virtual directory to return.

        Returns:
            The virtual directory from the IIS instance.
        """
        return VirtualDirectory(vdir_name, self)

    def get_webapp(self, app_name: str, /) -> WebApplication:
        """Get the specified web application from the IIS instance.

        Args:
            app_name: The name of the web application to return.

        Returns:
            The web application from the IIS instance.
        """
        return WebApplication(app_name, self)

    def get_webapp_pool(self, pool_name: str, /) -> WebApplicationPool:
        """Get the specified web application pool from the IIS instance.

        Args:
            pool_name: The name of the web application pool to return.

        Returns:
            The web application pool from the IIS instance.
        """
        return WebApplicationPool(pool_name, self)

    def get_website(self, site_name: str, /) -> WebSite:
        """Get the specified website from the IIS instance.

        Args:
            site_name: The name of the website to return.

        Returns:
            The website from the IIS instance.
        """
        return WebSite(site_name, self)

    def has_item(self, item_type: Type[IISObject], item_name: str, /) -> bool:
        """Determine if the specified item of the specified type exists in the IIS instance.

        Args:
            item_type: The type of the item for which to search.
            item_name: The name of the item for which to search.

        Returns:
            True if the item exists in the IIS instance, False otherwise.
        """
        return item_name in getattr(self, f'{self._IIS_TYPE_MAP[item_type]}s')

    def has_virtual_dir(self, vdir_name: str, /) -> bool:
        """Determine if the IIS instance has the specified virtual directory.

        Args:
            vdir_name: The name of the virtual directory for which to search.

        Returns:
            True if the virtual directory exists in the IIS instance, False otherwise.
        """
        return self.has_item(VirtualDirectory, vdir_name)

    def has_webapp(self, app_name: str, /) -> bool:
        """Determine if the IIS instance has the specified web application.

        Args:
            app_name: The name of the web application for which to search.

        Returns:
            True if the web application exists in the IIS instance, False otherwise.
        """
        return self.has_item(WebApplication, app_name)

    def has_webapp_pool(self, pool_name: str, /) -> bool:
        """Determine if the IIS instance has the specified web application pool.

        Args:
            pool_name: The name of the web application pool for which to search.

        Returns:
            True if the web application pool exists in the IIS instance, False otherwise.
        """
        return self.has_item(WebApplicationPool, pool_name)

    def has_website(self, site_name: str, /) -> bool:
        """Determine if the IIS instance has the specified website.

        Args:
            site_name: The name of the web application for which to search.

        Returns:
            True if the website exists in the IIS instance, False otherwise.
        """
        return self.has_item(WebSite, site_name)

    def manage_item(self, action: str, item_type: Type[IISObject], /, *args) -> CommandResult:
        """Manage an IIS object using the standard IIS appcmd.

        Args:
            action: The management action to perform.
            item_type: The type of the item to manage.
            *args (optional, default=[]): Any other appcmd arguments to pass to the management action.

        Returns:
            The result of the appcmd command.
        """
        return appcmd(action, self._IIS_TYPE_MAP[item_type], *args, hostname=self.hostname, remote_powershell=self._remote_powershell)

    def remove_webapp(self, appname: str, /) -> None:
        """Remove the specified web application from the IIS instance.

        Args:
            appname: The name of the web application to remove.

        Returns:
            Nothing.
        """
        self.manage_item('delete', WebApplication, appname)

    def remove_webapp_pool(self, pool_name: str, /) -> None:
        """Remove the specified web application pool from the IIS instance.

        Args:
            appname: The name of the web application pool to remove.

        Returns:
            Nothing.
        """
        self.manage_item('delete', WebApplicationPool, pool_name)

    def reset(self, verbose: bool = True) -> CommandResult:
        """Reset the IIS instance.

        Args:
            quiet (optional, default=False): If True do not print result to standard output stream.

        Returns:
            The result of the reset command.
        """
        return syscmd('iisreset', remote=self.hostname, show_stdout=not verbose)

    def start(self, quiet: bool = False) -> CommandResult:
        """Start the IIS instance.

        Args:
            quiet (optional, default=False): If True do not print result to standard output stream.

        Returns:
            The result of the start command.
        """
        return syscmd('iisreset', '/start', remote=self.hostname, show_stdout=not quiet)

    def stop(self, quiet: bool = False) -> CommandResult:
        """Stop the IIS instance.

        Args:
            quiet (optional, default=False): If True do not print result to standard output stream.

        Returns:
            The result of the stop command.
        """
        return syscmd('iisreset', '/stop', remote=self.hostname, show_stdout=not quiet)


class IISConfigurationSection:
    """Class to create a universal abstract interface for an IIS configuration section."""

    def __init__(self, name: str, /, path: Optional[PathName], set_location: Optional[str] = None,
                 hostname: Optional[str] = None, remote_powershell: Optional[bool] = None):
        """
        Args:
            name: The name of the IIS configuration section.
            path: The path to the configuration section.
            set_location (optional, default=None): If not None, will be applied to the appcmd /commit option.
            hostname (optional, default=localhost): The name of the IIS server hosting the instance.
            remote_powershell (optional): Determines if PowerShell remoting is used when executing appcmd against a remote server.
                Defaults to False for hostname is None, otherwise defaults to True.

        Attributes:
            _hostname: The value of the hostname argument.
            _name: The value of the name argument.
            _path: The value of the path argument.
            _remote_powershell: The resolved value of the remote_powershell argument.
            _set_location: The value of the set_location argument.
        """
        self._name = name
        self._path = path if path else ''
        self._set_location = set_location
        self._hostname = hostname
        self._remote_powershell = (True if (remote_powershell is None) else remote_powershell) if hostname else False

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __str__(self):
        return ''.join([line for line in self._run_appcmd('list', 'config', self._path, f'/section:{self._name}') if line.strip()])

    def __getattr__(self, attr: str) -> str:
        if (config := xmlparse_str(str(self))[0]).attrib['CONFIG.SECTION'] != self._name:
            raise IISConfigurationError(IISConfigurationError.PARSE_ERROR, expected=self._name)
        result = None
        if cfg_section := config.find(self._name.replace('/', '-')):
            if attr in cfg_section.attrib:
                return str_to_pythonval(cfg_section.attrib[attr])
            if attr.endswith('s'):
                if cfg_section_plural := cfg_section.findall('.//' + attr.rstrip('s'))[0]:
                    return str(cfg_section_plural.text)
            result = cfg_section.find(f'./[@{attr}]')
        if not result:
            raise AttributeError(attr)
        return str(result.text)

    def __setattr__(self, attr: str, value: str) -> None:
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return
        if isinstance(value, bool):
            value = bool_to_str(value)
        self._run_appcmd('set', 'config', self._path, f'/section:{self._name}', f'/{attr}:{value}')

    def _run_appcmd(self, *args) -> CommandResult:
        """Run the IIS appcmd against this IIS configuration section.

        Args:
            *args: The arguments to pass to appcmd.

        Returns:
            The output from appcmd.
        """
        cmd_args = list(args)
        if self._set_location:
            cmd_args += [f'/commit:{self._set_location}']
        return appcmd(*cmd_args, hostname=self._hostname, remote_powershell=self._remote_powershell)

    def add_collection_member(self, collection: str, /, properties: Dict, changes: Optional[Dict]) -> None:
        """Add the specified properties to the collection.

        Args:
            collection: The collection to which to add the properties.
            properties: The properties to add to the collection.
            changes: Changes to make to the properties before adding.

        Returns:
            Nothing.
        """
        self.add_property(collection, dict2expat((properties | changes) if changes else properties))

    def add_property(self, propname: str, value: str, /) -> None:
        """Add a property with the specified value.

        Args:
            propname: The name of the property to add.
            value: The value of the property to add.

        Returns:
            Nothing.
        """
        self._run_appcmd('set', 'config', self._path, f'/section:{self._name}', f'/+{propname}.{value}')

    def has_collection_member(self, collection: str, /, filter_on: str, value: str) -> bool:
        """Determine if a collection has a specified member.

        Args:
            collection: The collection to search.
            filter_on: Then member to filter on.
            value: The value for which to search.

        Returns:
            True if the collection has the specified member, False otherwise.
        """
        return bool([m for m in getattr(self, collection) if m.attrib[filter_on] == value])

    def rm_collection_member(self, collection: str, /, selectors: Dict) -> None:
        """Remove the specified properties from the collection.

        Args:
            collection: The collection from which to remove the properties.
            selectors: The selectors to identify the properties.

        Returns:
            Nothing.
        """
        self.rm_property(collection, dict2expat(selectors))

    def rm_property(self, propname: str, value: Optional[str] = None, /) -> None:
        """Remove a property conditionally with the specified value.

        Args:
            propname: The name of the property to remove.
            value (optional, default=None): If not None, only remove the property if it has the specified value.

        Returns:
            Nothing.
        """
        propspec = f'{propname}.{value}' if value else propname
        self._run_appcmd('set', 'config', self._path, f'/section:{self._name}', f'/-{propspec}')


class IISAdvancedLogger(IISConfigurationSection):
    """Class to create a universal abstract interface for the IIS advanced logger."""

    def __init__(self, path: Optional[PathName], /, logtype: str, set_location: str,
                 hostname: Optional[str] = None, remote_powershell: Optional[bool] = None):
        """
        Args:
            path: The path to AdvancedLogger configuration section.
            logtype: The AdvancedLogger type.
            set_location: Passed to the base class.
            hostname (optional, default=localhost): The name of the IIS server hosting the instance.
            remote_powershell (optional): Determines if PowerShell remoting is used when executing appcmd against a remote server.
                Defaults to False for hostname is None, otherwise defaults to True.
        """
        super().__init__(f'advancedLogging/{logtype}', path=path, set_location=set_location, hostname=hostname, remote_powershell=remote_powershell)

    def _run_appcmd(self, *cmd_args) -> CommandResult:
        """Run the IIS appcmd against this IIS advanced logger configuration.

        Args:
            *cmd_args: The arguments to pass to appcmd.

        Returns:
            The output from appcmd.
        """
        try:
            return super()._run_appcmd(*cmd_args)
        except IISConfigurationError as err:
            if ('message' not in err.vars) or ('Unknown config section "advancedLogging' not in err.vars['message']):
                raise
        raise IISAdvancedLogError(IISAdvancedLogError.NOT_INSTALLED)

    def add_field(self, field_id: str, field_values: Optional[Dict] = None, /) -> None:
        """Add a field to the advanced logger configuration.

        Args:
            field_id: The field ID to add.
            field_values (optional, default={}): Any field values to add to override the default values.

        Returns:
            Nothing.
        """
        default_values = {'id': field_id,
                          'sourceName': field_id,
                          'sourceType': 'RequestHeader',
                          'logHeaderName': field_id,
                          'category': 'Default',
                          'description': '',
                          'defaultValue': '',
                          'loggingDataType': 'TypeLPCSTR'}
        if field_values and 'sourceName' not in field_values:
            default_values['sourceName'] = field_id
        self.add_collection_member('fields', default_values, field_values)

    def add_log(self, log_name: str, /, log_values: Optional[Dict] = None, fields: Optional[Dict] = None) -> None:
        """Add a log definition to the advanced logger configuration.

        Args:
            log_name: The name of the log to add.
            log_values (optional, default=None): Any log values to add to override the default values.
            fields (optional, default=None): Any fields to add to the log definition.

        Returns:
            Nothing.
        """
        self.add_collection_member('logDefinitions',
                                   {'baseFileName': log_name,
                                    'writeLogDataToDisk': 'true',
                                    'enabled': 'true',
                                    'logRollOption': 'Schedule',
                                    'maxDurationSeconds': '86400',
                                    'maxFileSizeKB': '1024',
                                    'schedule': 'Hourly'},
                                   log_values)
        if fields:
            if not isinstance(fields, dict):
                fields = {f: {} for f in fields}
        else:
            fields = {}
        for (field, values) in fields.items():
            self.add_logfield(log_name, field, values)

    def add_logfield(self, log_name: str, field_name: str, field_values: Optional[Dict] = None, /) -> None:
        """Add a field to an advanced logger log definition.

        Args:
            log_name: The name of the log to which to add the field.
            field_name: The name of the field to add.
            field_values (optional, default=None): Any fields values to add to the field.

        Returns:
            Nothing.
        """
        if not self.has_collection_member('fields', 'id', field_name):
            raise IISAdvancedLogError(IISAdvancedLogError.BAD_FIELD, field=field_name)
        self.add_collection_member(f"logDefinitions.[baseFileName='{log_name}'].selectedFields",
                                   {'id': field_name,
                                    'logHeaderName': '',
                                    'required': 'false',
                                    'defaultValue': ''},
                                   field_values)

    def has_field(self, field_id: str, /) -> bool:
        """Determine if the advanced logger configuration has the specified field.

        Args:
            field_id: The field ID for which to search.

        Returns:
            True if the advanced logger configuration has the specified field, False otherwise.
        """
        return self.has_collection_member('fields', 'id', field_id)

    def rm_field(self, field_id: str, /) -> None:
        """Remove a field from the advanced logger configuration.

        Args:
            field_id: The field ID to remove.

        Returns:
            Nothing.
        """
        self.rm_collection_member('fields', {'id': field_id})

    def rm_log(self, log_name: str, /) -> None:
        """Remove a log definition to the advanced logger configuration.

        Args:
            log_name: The name of the log to remove.

        Returns:
            Nothing.
        """
        self.rm_collection_member('logDefinitions', {'baseFileName': log_name})


def appcmd(*cmd_args, hostname: Optional[str], **sys_cmd_args) -> CommandResult:
    """Interface to run the standard IIS appcmd command-line tool.

    Args:
        *cmd_args: The arguments to pass to appcmd.
        hostname: The hostname to pass to appcmd.
        **sys_cmd_args: The arguments to pass to syscmd when running appcmd.

    Returns:
        The result of the appcmd call.

    Raises:
        AppCmdError.APPCMD_ERROR: If there are errors reported by appcmd.
    """
    _appcmd = Path(getenv('SystemRoot', ''), 'system32/inetsrv/appcmd.exe')

    if hostname and ('remote_powershell' not in sys_cmd_args):
        sys_cmd_args['remote_powershell'] = True
    try:
        return syscmd(str(_appcmd), *cmd_args, '/xml', remote=hostname, **sys_cmd_args)
    except CMDError as err:
        if ('outlines' not in err.vars) or not err.vars['outlines']:
            raise

        if (errmsg := xmlparse_list(err.vars['outlines']).find('ERROR')) is None:
            raise
        return_code = err.vars['returncode']

    err_object = AppCmdError(AppCmdError.APPCMD_ERROR, message=errmsg.attrib['message'])
    err_object.vars['returncode'] = return_code
    raise err_object


def dict2expat(py_dict: Dict, /) -> str:
    """Converts Python dictionaries to the syntax understood by the IIS appcmd command-line tool.

    Args:
        py_dict: The Python dictionary to convert.

    Returns:
        The appcmd format for the dictionary.
    """
    return '[' + ','.join([f"{k}='{v}'" for k, v in py_dict.items()]) + "]"

# cSpell:ignore iisreset inetsrv syscmd vdir



"""This module provides utilities for building automation."""

# Import standard modules
import sys
from abc import abstractmethod
from logging import Logger
from pathlib import Path
from typing import Any, Callable, Dict, Optional

# Import internal modules
from .lang import CommandResult
from .sysutil import popd, SysCmdRunner


class ActionCommandRunner(SysCmdRunner):  # pylint: disable=too-few-public-methods
    """Class to wrap SysCmdRunner for simple usage with auto-logging."""

    def __init__(self, command: str, /, *args, logger: Optional[Callable | Logger] = print, guard: str = '', syscmd_args: Optional[Dict[Any, Any]] = None, **kwargs: Any):
        """
        Args:
            command: The command passed to SysCmdRunner.
            logger (optional, default=print): A logging instance to use when the command is run.
            guard (optional, default=''): A line to be printed before the command is run.
                If an empty string, nothing is printed.
            syscmd_args (optional, default={}): Any arguments passed to syscmd.
            *args (optional, default=()): The list of default args passed to SysCmdRunner.
            **kwargs (optional, default={}): The list of default named args passed to SysCmdRunner.

        Attributes:
            logger: The value of the logger argument.
            guard: The value of the guard argument.
        """
        super().__init__(command, *args, syscmd_args=syscmd_args, **kwargs)
        self.logger = logger.info if isinstance(logger, Logger) else logger
        self.guard = guard

    def run(self, message: str, *args, post_option_args: Optional[Dict] = None,  # type: ignore[override]  # pylint: disable=arguments-differ
            syscmd_args: Optional[Dict[Any, Any]] = None, **kwargs) -> CommandResult:
        """Run the action.

        Args:
            message: The message to log if a logger has been set.
            post_option_args (optional, default=[]): The list of post_option_args passed to SysCmdRunner.
            syscmd_args (optional, default={}): The syscmd_args passed to SysCmdRunner.
            *args (optional, default=[]): The list of args passed to SysCmdRunner.
            **kwargs (optional, default={}): The list of named args passed to SysCmdRunner.

        Returns:
            Returns whatever is returned by SysCmdRunner.
        """
        if self.logger and message:
            if self.guard:
                self.logger(self.guard)
            self.logger(message)
        return super().run(*args, post_option_args=post_option_args, syscmd_args=syscmd_args, **kwargs)


class Action:
    """The common base class for all actions.

    This is a virtual class and the inheriting class must at least include a _execute() method.

    The action is invoked by calling the execute() method which will run the following methods::

        pre()
        _execute()
        post()

    These are run in a try to catch any exceptions with the always_post() method run in the finally block.

    Attributes:
        message_guard: This string is printed by logger if the value of guard passed to logger is true.
    """
    message_guard = f"{'*'*70}"

    def __init__(self, **_unused_kwargs: Any):
        """
        Args:
            **_unused_kwargs: All arguments passed are ignored by the time they reach this initializer.

        Attributes:
            _project_root: The parent directory from where the current action is run.
        """
        self._project_root = Path.cwd().parent

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    @abstractmethod
    def _execute(self) -> None:
        pass

    project_root = property(lambda s: s._project_root, doc='A read-only property which returns the root of the automation.')

    def pre(self) -> None:
        """Executed before _execute().

        Returns:
            Nothing.
        """

    def post(self) -> None:
        """Executed after _execute().

        Returns:
            Nothing.
        """

    def always_post(self) -> None:
        """Always executed after _execute() as in finally in try/catch/finally.

        Returns:
            Nothing.
        """
        popd()

    def execute(self) -> None:
        """Run the _execute() method from the child class.

        Returns:
            Nothing.
        """
        try:
            self.pre()
            self._execute()
            self.post()
        finally:
            self.always_post()

    def log_message(self, message: str, /, guard: bool = False, leader: str = 'INFO') -> None:
        """Log a message to stdout and flushes the stream.

        Args:
            message: The message to be printed.
            guard (optional, default=False): If True, message_guard will be printed on a line before the message.
            leader (optional, default='INFO'): If it does not evaluate to False, it will be prepended to every printed
                line, including the guard.

        Returns:
            Nothing.
        """
        the_leader: str = f'{leader} ' if leader else ''
        if the_guard := (self.message_guard if (guard is True) else ''):
            print(f'{the_leader}{the_guard}')
        print(f'{the_leader}{message}')
        sys.stdout.flush()

# cSpell:ignore syscmd



"""Unit tests for the sysutil module."""

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name
# flake8: noqa

from enum import Enum
from multiprocessing import Process, Queue
from os import fdopen
from pathlib import Path
from tempfile import mkdtemp, mkstemp
from unittest import main, skip, TestCase

from batcave.sysutil import pushd, popd, CMDError, LockFile, LockError, LockMode

LockSignal = Enum('LockSignal', ('true', 'false'))


class TestExceptions(TestCase):
    def test_PlatformException(self):
        try:
            raise CMDError(CMDError.INVALID_OPERATION, platform='badplatform')
        except CMDError as err:
            self.assertEqual(CMDError.INVALID_OPERATION.code, err.code)


class TestLockFile(TestCase):
    def setUp(self):
        (fd, fn) = mkstemp()
        self._fh = fdopen(fd)
        self._fn = Path(fn)
        self._gotlock = Queue()
        self._lockagain = Process(target=secondary_lock_process, args=(self._fn, self._gotlock))

    def tearDown(self):
        self._gotlock.close()
        self._fh.close()
        if self._fn.exists():
            self._fn.unlink()

    def test_cleanup(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True):
            pass
        self.assertFalse(self._fn.exists())

    def test_no_cleanup(self):
        with LockFile(self._fn, handle=self._fh, cleanup=False):
            pass
        self.assertTrue(self._fn.exists())

    @skip('Problems with secondary process')
    def test_lock(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True):
            self._lockagain.start()
            gotlock = self._gotlock.get()
            self._lockagain.join()
            self.assertTrue(gotlock == LockSignal.false)

    @skip('Problems with secondary process')
    def test_unlock(self):
        with LockFile(self._fn, handle=self._fh, cleanup=True) as lockfile:
            lockfile.action(LockMode.unlock)
            self._lockagain.start()
            gotlock = self._gotlock.get()
            self._lockagain.join()
            lockfile.action(LockMode.lock)
            self.assertTrue(gotlock == LockSignal.true)


def secondary_lock_process(filename, queue):
    try:
        with LockFile(filename, cleanup=False) as lockagain:
            lockagain.action(LockMode.unlock)
            queue.put(LockSignal.true)
    except LockError:
        queue.put(LockSignal.false)


class TestDirStack(TestCase):
    def setUp(self):
        self._tempdir = Path(mkdtemp())

    def tearDown(self):
        self._tempdir.rmdir()

    def test_push_and_pop(self):
        start = Path.cwd()

        olddir = pushd(self._tempdir)
        newdir = Path.cwd()
        self.assertEqual(olddir, start)
        self.assertEqual(newdir, self._tempdir)

        olddir = popd()
        newdir = Path.cwd()
        self.assertEqual(olddir, start)
        self.assertEqual(newdir, start)


if __name__ == '__main__':
    main()



"""This module provides a Pythonic interface to the QuickBuild RESTful API."""

# Import standard modules
from typing import cast, Any, Dict, List, Optional, Union
from xml.etree.ElementTree import fromstring, tostring, Element

# Import third-party modules
from requests import codes, delete as req_del, get as req_get, post as req_post, Response
from requests.exceptions import HTTPError

# Import internal modules
from .lang import bool_to_str


class QuickBuildObject:
    """Class to create a universal abstract interface for a QuickBuild object."""

    def __init__(self, console: 'QuickBuildConsole', object_id: int | str, /):
        """
        Args:
            console: The QuickBuild console containing this object.
            object_id: The object ID.

        Attributes:
            _console: The value of the console argument.
            _object_id: The value of the object_id argument.
            _object_path: The RESTful API path to the object.
            _object_type: The object type.
        """
        self._console = console
        self._object_id = int(object_id)
        object_name = self.__class__.__name__.lower().replace('quickbuild', '')
        self._object_type = 'configuration' if (object_name == 'cfg') else f'{object_name}'
        self._object_path = f'{self._object_type}s/{self._object_id}'

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> Optional[str]:
        try:
            return self._console.qb_runner(f'{self._object_path}/{attr}').text
        except HTTPError as err:
            if not err.response.status_code == 500:
                raise
        raise AttributeError(f'{self._object_type.capitalize()} {self._object_id} has no attribute: {attr}')

    def __setattr__(self, attr: str, value: str) -> None:
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return

        getattr(self, attr)  # Need to raise AttributeError if this attribute doesn't exist
        object_xml = fromstring(self._console.qb_runner(self._object_path).text)
        attr_element = cast(Element, object_xml.find(attr))
        attr_element.text = value
        self._console.qb_runner(f'{self._object_type}s', xml_data=object_xml)

    def __str__(self):
        return self._console.qb_runner(self._object_path).text

    id = property(lambda s: s._object_id, doc='A read-only property which returns the QuickBuild object ID.')


class QuickBuildBuild(QuickBuildObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a QuickBuild configuration run."""


class QuickBuildDashboard(QuickBuildObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a QuickBuild dashboard."""


class QuickBuildCfg(QuickBuildObject):
    """Class to create a universal abstract interface for a QuickBuild configuration."""

    def _get_id(self, thing: Union[int, 'QuickBuildCfg'], /) -> int:
        """Get the ID of the specified configuration.

        Args:
            thing: The configuration for which to return the ID.

        Returns:
            The ID of the specified configuration.
        """
        if isinstance(thing, int):
            return thing
        if isinstance(thing, QuickBuildCfg):
            return thing.id
        return self._console.configs[thing].id

    @property
    def latest_build(self) -> QuickBuildBuild:
        """A read-only property which returns the latest build for this configuration."""
        build_element = str(cast(Element, fromstring(self._console.qb_runner(f'latest_builds/{self._object_id}').text).find('id')).text)
        return QuickBuildBuild(self._console, build_element)

    def change_var(self, var: str, val: str, /) -> 'QuickBuildCfg':
        """Change the value of a variable in this configuration.

        Args:
            var: The variable to change.
            val: The new value of the variable.

        Returns:
            The configuration.
        """
        cfg_xml = fromstring(self._console.qb_runner(f'configurations/{self._object_id}').text)
        for var_xml in cast(Element, cfg_xml.find('variables')).findall('com.pmease.quickbuild.variable.Variable'):
            if cast(Element, var_xml.find('name')).text == var:
                val_ref = cast(Element, cast(Element, var_xml.find('valueProvider')).find('value'))
                val_ref.text = val
        self._console.qb_runner('configurations', xml_data=cfg_xml)
        return self

    def copy(self, parent: 'QuickBuildCfg', name: str, /, *, recurse: bool = False) -> 'QuickBuildCfg':
        """Copy this configuration to a new configuration.

        Args:
            parent: The parent to which to copy the configuration.
            name: The name of the new configuration.
            recurse (optional, default=False): If True, copy the children configurations also.

        Returns:
            The new configuration.
        """
        new_id = self._console.qb_runner(f'configurations/{self._object_id}/copy?parent_id={self._get_id(parent)}&name={name}&recursive=' + bool_to_str(recurse))
        new_cfg = QuickBuildCfg(self._console, str(new_id.text))
        self._console.updater()
        return new_cfg

    def disable(self, /, *, wait: bool = False) -> None:
        """Disable this configuration.

        Args:
            wait (optional, default=False): If True, wait for the configuration to be disabled.

        Returns:
            Nothing.
        """
        self.disabled = 'true'  # pylint: disable=attribute-defined-outside-init
        while wait and str(self.latest_build.status).upper() not in ('SUCCESSFUL', 'FAILED'):
            pass

    def enable(self) -> None:
        """Enable this configuration.

        Returns:
            Nothing.
        """
        self.disabled = 'false'  # pylint: disable=attribute-defined-outside-init

    def get_children(self, /, *, recurse: bool = False) -> List['QuickBuildCfg']:
        """Get the child configurations.

        Args:
            recurse (optional, default=False): Get the children recursively.

        Returns:
            The list of child configurations.
        """
        ans = self._console.qb_runner(f'configurations?parent_id={self._object_id}&recursive=' + bool_to_str(recurse))
        return [QuickBuildCfg(self._console, str(cast(Element, c.find('id')).text)) for c in fromstring(ans.text).findall('com.pmease.quickbuild.model.Configuration')]

    children = property(get_children, doc='A read-only property which returns a list of children for the object.')

    def remove(self) -> None:
        """Remove this configuration.

        Returns:
            Nothing.
        """
        self._console.qb_runner(f'configurations/{self._object_id}', delete=True)

    def rename(self, newname: str, /) -> 'QuickBuildCfg':
        """Rename this configuration.

        Args:
            newname: The new name for the configuration.

        Returns:
            The renamed configuration.
        """
        cfg_xml = fromstring(self._console.qb_runner(f'configurations/{self._object_id}').text)
        name = cast(Element, cfg_xml.find('name'))
        name.text = newname
        self._console.qb_runner('configurations', xml_data=cfg_xml)
        self._console.updater()
        return self

    def reparent(self, new_parent: 'QuickBuildCfg', /, *, rename: bool = False) -> 'QuickBuildCfg':
        """Reparent this configuration.

        Args:
            new_parent: The new parent for the configuration.
            rename (optional, default=False): If not False, rename the configuration when moved.

        Returns:
            The moved configuration.
        """
        cfg_xml = fromstring(self._console.qb_runner(f'configurations/{self._object_id}').text)
        parent = cast(Element, cfg_xml.find('parent'))
        parent.text = str(new_parent.id)
        if rename:
            name = cast(Element, cfg_xml.find('name'))
            name.text = bool_to_str(rename)
        self._console.qb_runner('configurations', xml_data=cfg_xml)
        self._console.updater()
        return self


class QuickBuildConsole:
    """Class to create a universal abstract interface for a QuickBuild console."""

    def __init__(self, host: str, /, *, user: str, password: str):
        """
        Args:
            host: The server hosting the QuickBuild console.
            user: The QuickBuild user for API access.
            password: The QuickBuild password for API access.

        Attributes:
            _host: The value of the host argument.
            _password: The value of the password argument.
            _update: When True, the internal values need to be refreshed from the API.
            _user: The value of the user argument.
        """
        self._host = host
        self._user = user
        self._password = password
        self._update = True
        self.configs: Dict[str, QuickBuildCfg] = {}
        self.dashboards: Dict[str, QuickBuildDashboard] = {}
        self.updater()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> QuickBuildCfg:
        if attr in self.configs:
            return self.configs[attr]
        raise AttributeError(f'No such configuration: {attr}')

    @property
    def update(self) -> bool:
        """A read-write property which returns and sets the update state for the console."""
        return self._update

    @update.setter
    def update(self, val: bool) -> None:
        self._update = val
        self.updater()

    def create_dashboard(self, name: str, dashboard: str | QuickBuildDashboard, /) -> QuickBuildDashboard:
        """Create a dashboard from an existing one.

        Args:
            name: The name of the new dashboard.
            dashboard: The dashboard from which to make the copy.

        Returns:
            The new dashboard.
        """
        xml_data: str | Element = str(dashboard) if isinstance(dashboard, QuickBuildDashboard) else dashboard
        if isinstance(xml_data, str):
            xml_data = fromstring(xml_data)
        if (id_tag := xml_data.find('id')) is not None:
            xml_data.remove(id_tag)
        cast(Element, xml_data.find('name')).text = name
        return QuickBuildDashboard(self, str(self.qb_runner('dashboards', xml_data=xml_data).text))

    def get_dashboard(self, dashboard: str, /) -> QuickBuildDashboard:
        """Get the named dashboard.

        Args:
            dashboard: The dashboard to return.

        Returns:
            The requested dashboard.
        """
        return self.dashboards[dashboard]

    def has_dashboard(self, dashboard: str, /) -> bool:
        """Determine if the specified dashboard exists.

        Args:
            dashboard: The dashboard for which to search.

        Returns:
            True if the requested dashboard exists, False otherwise.
        """
        return dashboard in self.dashboards

    def qb_runner(self, cmd: str, /, *, xml_data: Optional[Any] = None, delete: bool = False) -> Response:
        """Provide an interface to the RESTful API.

        Args:
            cmd: The API command to run.
            xml_data (optional, default=None): Any data to pass to the command.
            delete (optional, default=False): If True, use delete, otherwise use get.

        Returns:
            The result of the API call.
        """
        caller: Any = None
        api_call = f'http://{self._host}/rest/{cmd}'
        api_args: Dict[str, Any] = {'auth': (self._user, self._password)}
        if delete:
            caller = req_del
        elif xml_data is None:
            caller = req_get
        else:
            caller = req_post
            api_args['data'] = xml_data if isinstance(xml_data, str) else tostring(xml_data)
        if (result := caller(api_call, **api_args)).status_code != codes.ok:  # pylint: disable=no-member
            result.raise_for_status()
        return result

    def updater(self) -> None:
        """Update the configuration list.

        Returns:
            Nothing.
        """
        if self._update:
            top = QuickBuildCfg(self, 1)
            self.configs = {str(c.path): c for c in top.get_children(recurse=True)}
            self.configs[str(top.path)] = top

            for dashboard in fromstring(self.qb_runner('dashboards').text).iter('com.pmease.quickbuild.model.Dashboard'):
                self.dashboards[str(cast(Element, dashboard.find('name')).text)] = QuickBuildDashboard(self, str(cast(Element, dashboard.find('id')).text))

# cSpell:ignore tful quickbuild



"""This module implements a simple state machine."""

# Import standard modules
from enum import Enum
from pathlib import Path
from string import Template
from typing import Sequence

# Import internal modules
from .sysutil import LockFile
from .lang import DEFAULT_ENCODING, is_debug, BatCaveError, BatCaveException, PathName

StateStatus = Enum('StateStatus', ('entering', 'exited'))


class StateMachineError(BatCaveException):
    """State Machine Exceptions.

    Attributes:
        ALREADY_STARTED: There was an attempt to start a state machine that is already in progress.
        BAD_ENTRY: There was an attempt to enter next state before exiting current one.
        BAD_EXIT: There was an attempt to exit the state before it was entered.
        BAD_ROLLBACK: There was an attempt to rollback a state before it entered.
        BAD_STATUS: An invalid status was requested.
        CRASHED: The state machine crashed in a state.
        DONE: There was an attempt to enter a state after the final state.
        NOT_STARTED: There was an attempt to continue a state machine that has not started.
    """
    ALREADY_STARTED = BatCaveError(1, 'State machine already started')
    BAD_ENTRY = BatCaveError(2, 'Attempt to enter next state before exiting current one')
    BAD_EXIT = BatCaveError(3, 'Attempt to exit state before entering')
    BAD_ROLLBACK = BatCaveError(4, 'Attempt to rollback state before entering')
    BAD_STATUS = BatCaveError(5, Template('Unknown status: $status'))
    CRASHED = BatCaveError(6, Template('State machine crashed in state: $state'))
    DONE = BatCaveError(7, 'Attempt to enter state after final state')
    NOT_STARTED = BatCaveError(8, 'State machine not started')


class StateMachine:
    """This class implements a classic state machine.

    Attributes:
        _DEFAULT_STATEFILE: The default file to use to store the current state information.
        _DEFAULT_LOGFILE: The default file to use to store the log information.
        _DEFAULT_LOCKFILE: The default file to use for file locking.
    """
    _DEFAULT_LOCKFILE = Path('lock')
    _DEFAULT_STATEFILE = Path('state')

    def __init__(self, states: Sequence, /, *, statefile: PathName = _DEFAULT_STATEFILE, lockfile: PathName = _DEFAULT_LOCKFILE, autostart: bool = True):
        """
        Args:
            states: The list of states for the state machine.
            statefile (optional, default=_DEFAULT_STATEFILE): The value of the file in which to store the state machine state.
            lockfile (optional, default=_DEFAULT_LOCKFILE): The file to use to lock the state machine.
            autostart (optional, default=True): If True, start the state machine when the instance is created.

        Attributes:
            state: Indicates the current state of the state machine.
            status: Indicates the current status of the state machine.
            _locker: The value of the lockfile argument.
            _started: Indicates if the state machine is running.
            _statefile: The value of the statefile argument.
            _states: The value of the states argument prepended by 'None.'

        Raises:
            StateMachineError.BAD_STATUS: if the value of self.status is not in StateStatus
        """
        self._statefile = Path(statefile)
        self._locker = LockFile(lockfile)
        self._states = ['None'] + list(states)
        self._started = False
        if self._statefile.exists():
            debug_msg = 'Found'
            with open(self._statefile, encoding=DEFAULT_ENCODING) as filestream:
                (status, state) = filestream.read().split()
                self.status = StateStatus[status]
                self.state = state
                if self.status not in StateStatus:
                    raise StateMachineError(StateMachineError.BAD_STATUS, status=self.status)
        else:
            debug_msg = 'Initializing'
            self.reset()

        if autostart:
            self.start()

        if is_debug('STATEMACHINE'):
            print(f'{debug_msg} state file with {self.status.name} {self.state}')

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.done()
        return False

    def _writestate(self) -> None:
        """Write the current state to the state file.

        Returns:
            Nothing.
        """
        if is_debug('STATEMACHINE'):
            print(f'Writing state file with {self.status.name} {self.state}')
        with open(self._statefile, 'w', encoding=DEFAULT_ENCODING) as filestream:
            print(self.status.name, self.state, file=filestream)

    def done(self) -> None:
        """Shutdown the state machine.

        Returns:
            Nothing.
        """
        self._locker.close()

    def enter_next_state(self) -> None:
        """Enter the next state.

        Returns:
            Nothing.

        Raises:
            StateMachineError.BAD_ENTRY: If the state machine has not exitted from the previous start.
            StateMachineError.DONE: If the state machine has already completed.
            StateMachineError.NOT_STARTED: If the state machine has not been started.
        """
        if not self._started:
            raise StateMachineError(StateMachineError.NOT_STARTED)
        if self.status != StateStatus.exited:
            raise StateMachineError(StateMachineError.BAD_ENTRY)
        if (next_state_index := self._states.index(self.state) + 1) >= len(self._states):
            raise StateMachineError(StateMachineError.DONE)
        self.status = StateStatus.entering
        self.state = self._states[next_state_index]
        self._writestate()

    def exit_state(self) -> None:
        """Exit the current state.

        Returns:
            Nothing.

        Raises:
            StateMachineError.BAD_EXIT: If the state machine has not entered a state.
            StateMachineError.NOT_STARTED: If the state machine has not been started.
        """
        if not self._started:
            raise StateMachineError(StateMachineError.NOT_STARTED)
        if self.status != StateStatus.entering:
            raise StateMachineError(StateMachineError.BAD_EXIT)
        self.status = StateStatus.exited
        self._writestate()

    def reset(self) -> None:
        """Reset the state machine.

        Returns:
            Nothing.
        """
        if self._statefile.exists():
            self._statefile.unlink()
        (self.status, self.state) = (StateStatus.exited, self._states[0])
        self._writestate()

    def rollback(self) -> None:
        """Rollback to the previous state.

        Returns:
            Nothing.

        Raises:
            StateMachineError.BAD_ROLLBACK: If the state machine has not entered a state.
        """
        if self.status != StateStatus.entering:
            raise StateMachineError(StateMachineError.BAD_ROLLBACK)
        self.status = StateStatus.exited
        self.state = self._states[self._states.index(self.state) - 1]
        self._writestate()

    def start(self) -> None:
        """Start the state machine.

        Returns:
            Nothing.

        Raises:
            StateMachineError.CRASHED: If the state machine crashed on the previous run.
            StateMachineError.ALREADY_STARTED: If the state machine has already been started.
        """
        if self._started:
            raise StateMachineError(StateMachineError.ALREADY_STARTED)
        if self.status == StateStatus.entering:
            raise StateMachineError(StateMachineError.CRASHED, state=self.state)
        self._started = True



"""This module provides utilities for managing file expansions.

An instantiation of this class will convert an XML into the procedure object::

    <procedure schema = "{schema-version}" >
        <header>HEADER</header>

        <flags>
            <FlagName question="Some Question">True|False, Yes|No,  1,0</FlagName>
        </flags>

        <environments>
            <common>
                <VariableName1>Variable Value 1</VariableName1>
            </common>
            <Environment1>
                <VariableName2>Variable Value 2</VariableName1>
            <Environment2
        </environments>

        <directories>
            <directory>Directory - Path</directory>
        </directories>

        <steps>
            <step condition="Optional Condition Expression" vars="var1=val1, var2=val2", repeat="var=val1, val2">
                "Multi-Step Text"
                <step>Step Text</step>
                <step import="Library Step Name">Alternate Step Text</step>
            </step>
        </steps>

        <step-library>
            <step name="Import Name" />
        </step-library>
    </procedure>
"""

# Import standard modules
from collections import OrderedDict as odict
from copy import deepcopy
from enum import Enum
from os import walk
from pathlib import Path
from re import compile as re_compile
from shutil import copyfile
from string import Template
from typing import cast, Any, Dict, List, Match, Optional, Sequence, Tuple
from xml.etree.ElementTree import fromstringlist as xml_parse, Element

# Import internal modules
from .fileutil import slurp
from .lang import DEFAULT_ENCODING, is_debug, str_to_pythonval, BatCaveError, BatCaveException, PathName, WIN32

OutputFormat = Enum('OutputFormat', ('text', 'html', 'csv'))


class ExpanderError(BatCaveException):
    """Expansion Exceptions.

    Attributes:
        NO_POST_DELIMITER: No postfix delimiter was found.
        NO_REPLACEMENT: No replacement was found for the specified variable
    """
    NO_POST_DELIMITER = BatCaveError(1, Template('No closing delimiter found in $substr'))
    NO_REPLACEMENT = BatCaveError(2, Template('No replacement found for variable ($var) in: $thing'))


class ProcedureError(BatCaveException):
    """Procedure Exceptions.

    Attributes:
        BAD_ENVIRONMENT: The requested environment is not valid.
        BAD_FLAG: The specified flag value is not value.
        BAD_FORMAT: The requested output format is not valid.
        BAD_LIBRARY: The requested library is not valid.
        BAD_SCHEMA: The procedure schema is not supported.
        EXPANSION_ERROR: The was an unspecified error during expansion.
    """
    BAD_ENVIRONMENT = BatCaveError(1, Template('Unknown environment requested: $env.'))
    BAD_FLAG = BatCaveError(2, Template('Invalid value for flag: $value'))
    BAD_FORMAT = BatCaveError(3, Template('Unknown output format requested: $format'))
    BAD_LIBRARY = BatCaveError(4, Template('Unable to locate import: $lib.'))
    BAD_SCHEMA = BatCaveError(5, Template('Procedure specified in wrong schema ($schema). Please use schema $expected.'))
    EXPANSION_ERROR = BatCaveError(6, Template('$err\n  On line: $text'))


class Formatter:
    """Class to hold formatting information.

    Attributes:
        _LINK_PRELIM: The prefix to indicate a hyperlink during formatting.
    """
    _LINK_PRELIM = '{link:'

    def __init__(self, output_format: OutputFormat, /):
        """
        Args:
            output_format: The output format.

        Attributes:
            format: The value of the name argument.
            count: This current character count on the line.
            keeper: The indentation stack
            level: The current indentation level.
            link_regex: The regular expression to locate a hyperlink.
            prefix: The line prefix for the current line.

        Raises:
            ProcedureError.BAD_FORMAT: If the requested output format is invalid.
        """
        self.format = output_format
        self.level = 0
        self.count = 1
        self.keeper: List[Tuple[int, str]] = []
        self.prefix = ''
        self.link_regex = re_compile(f'\\{self._LINK_PRELIM}(.+?)(\\|(.+))?\\}}')

        match self.format:
            case OutputFormat.csv | OutputFormat.text:
                self.bos = self.eos = self._bol = ''
                self.eol = '\n'
            case OutputFormat.html:
                self.bos = '<ul>'
                self.eos = '</ul>'
                self._bol = '<h2><li>'
                self.eol = '</li></h2>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)

    @property
    def bol(self) -> str:
        """A read-only property which returns the beginning of line formatting."""
        if self.level == 0:
            sep = ',' if (self.format == OutputFormat.csv) else '. '
            match self.format:
                case OutputFormat.csv | OutputFormat.text:
                    return chr(64 + self.count) + sep
                case OutputFormat.html:
                    return self._bol
                case _:
                    raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        else:
            if self.format == OutputFormat.csv:
                space = ''
                sep = ','
            else:
                space = '    ' * self.level
                sep = ': '
            return f'{self._bol}{space}{self.prefix}{self.count}{sep}'
        raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)

    def format_hyperlinks(self, line: str, /) -> str:
        """Format the hyperlinks in a line.

        Args:
            line: The line for which to format hyperlinks.

        Returns:
            The formatted line.

        Raises:
            ProcedureError.BAD_FORMAT: If the requested format is not valid.
        """
        if not line or (self._LINK_PRELIM not in line):
            return line

        if not (match := self.link_regex.search(line)):
            raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        replace_what = match.group(0)
        link = match.group(1)
        text = match.group(3) if (len(match.groups()) == 3) else ''
        replace_with = ''
        match self.format:
            case OutputFormat.csv:
                replace_with = f'"=HYPERLINK(""{link}"", ""{text}"")"' if text else link
            case OutputFormat.text:
                replace_with = f'{text} ({link})' if text else link
            case OutputFormat.html:
                text = text if text else link
                replace_with = f'<a href="{link}">{text}</a>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        return line.replace(replace_what, replace_with)

    def increment(self) -> None:
        """Increment the counter at the current indentation level.

        Returns:
            Nothing.
        """
        self.count += 1

    def indent(self) -> None:
        """Increment the indentation level.

        Returns:
            Nothing.
        """
        self.keeper.append((self.count, self.prefix))
        if self.level > 0:
            self.prefix = f'{self.keeper[-1][1]}{self.count}.'
        elif self.format == OutputFormat.html:
            self._bol = '<li style="color:white"><span style="color:black">'
            self.eol = '</span></li>'
        self.level += 1
        self.count = 1

    def outdent(self) -> None:
        """Decrement the indentation level.

        Returns:
            Nothing.
        """
        self.level -= 1
        (self.count, self.prefix) = self.keeper.pop()
        if self.level > 0:
            self.count += 1
        elif self.format == OutputFormat.html:
            self._bol = '<h2><li>'
            self.eol = '</li></h2>'


class Expander:
    """Class to handle interpolation of strings and files.

    Attributes:
        _PRELIM_DEFAULT: The prefix for a variable to be expanded.
        _POSTLIM_DEFAULT: The suffix for a variable to be expanded.
    """
    _PRELIM_DEFAULT = '{var:'
    _POSTLIM_DEFAULT = '}'

    def __init__(self, *, var_dict: Optional[Dict[str, str]] = None, var_props: Any = None, prelim: str = _PRELIM_DEFAULT, postlim: str = _POSTLIM_DEFAULT):
        """
        Args:
            var_dict: A dictionary of expansion variables.
            var_prop: An object with properties that resolve variables.
            prelim: The leading string to identify an expansion.
            postlim: The training string to identify an expansion.

        Attributes:
            postlim: The value of the postlim argument.
            prelim: The value of the prelim argument.
            re_var: The regular expression to identify a variable to expand.
            var_dict: The value of the var_dict argument.
            var_prop: The value of the var_prop argument.
        """
        self.var_dict = var_dict if var_dict else {}
        self.var_props = var_props if (isinstance(var_props, (list, tuple))) else [var_props]
        self.prelim = prelim
        self.postlim = postlim
        prelim_re = self.prelim
        postlim_re = self.postlim
        for spec in '.^$*+?|!/{}[]()<>:':
            prelim_re = prelim_re.replace(spec, '\\' + spec)
            postlim_re = postlim_re.replace(spec, '\\' + spec)
        self.re_var = re_compile(f'{prelim_re}([.a-zA-Z0-9_:]+){postlim_re}')

    def evaluate_expression(self, expression: str, /) -> bool:
        """Evaluate an expression in the expansion.

        Args:
            expression: The expression to evaluate.

        Returns:
            The evaluated expression.

        Raises:
            ExpanderError.NO_REPLACEMENT: If no replacement was found for the requested expansion variable.
        """
        if is_debug('EXPANDER'):
            print(f'Expanding (raw expression) "{expression}"')

        if isinstance(expression, bool):
            return expression
        if not expression:
            return True

        expression = self.expand(expression)
        synonyms = {' not ': ('!', '-not'),
                    ' and ': ('&&', ',', '-and'),
                    ' or ': ('||', '-or')}
        for (operator, syn_list) in synonyms.items():
            for synonym in syn_list:
                expression = expression.replace(synonym, operator)
        try:
            if is_debug('EXPANDER'):
                print(f'Expanding (corrected expression) "{expression}"')
            result = eval(expression, self.var_dict)  # pylint: disable=eval-used
            if is_debug('EXPANDER'):
                print(f'Expanding (evaluated expression) "{result}"')
            if isinstance(result, str):
                result = str_to_pythonval(result)
            if is_debug('EXPANDER'):
                print(f'Returning: "{result}"')
        except NameError as err:
            raise ExpanderError(ExpanderError.NO_REPLACEMENT, var=str(err), thing=expression) from err
        return result

    def expand(self, thing: Any, /) -> Any:
        """Perform an expansion on a Python object.

        Args:
            thing: The object on which to perform an expansion.

        Returns:
            The object with expansions performed.

        Raises:
            ExpanderError.NO_POST_DELIMITER: If the closing delimiter was not found.
            ExpanderError.NO_REPLACEMENT: If no replacement was found for the requested expansion variable.
        """
        if isinstance(thing, tuple):
            return {self.expand(i) for i in thing}

        if isinstance(thing, list):
            return [self.expand(i) for i in thing]

        if isinstance(thing, dict):
            return {self.expand(k): self.expand(v) for (k, v) in thing.items()}

        if (thing is None) or isinstance(thing, bool):
            return thing

        var = substr = replacer = ''
        while self.prelim in thing:
            try:
                var = cast(Match[str], self.re_var.search(thing)).group(1)
            except AttributeError as err:
                prelim_index = thing.index(self.prelim)
                substr = thing[prelim_index:prelim_index + 200]
                raise ExpanderError(ExpanderError.NO_POST_DELIMITER, substr=substr) from err

            fail = True
            if var in self.var_dict:
                replacer = self.var_dict[var]
                fail = False
            else:
                for var_props in self.var_props:
                    if hasattr(var_props, var):
                        replacer = getattr(var_props, var)
                        fail = False
                        break

            if fail:
                raise ExpanderError(ExpanderError.NO_REPLACEMENT, var=var, thing=thing)

            thing = thing.replace(f'{self.prelim}{var}{self.postlim}', str(replacer))
        return thing

    def expand_directory(self, source_dir: PathName, /, target_dir: Optional[PathName] = None, *,
                         ignore_files: Sequence[str] = tuple(), no_expand_files: Sequence[str] = tuple(), err_if_exists: bool = True) -> None:
        """Perform an expansion on an entire directory tree.

        Args:
            source_dir: The name of the directory on which to perform the expansions.
            target_dir (optional, default=None): The name of the output directory for the expansion results if not None,
                otherwise the current directory is used.
            ignore_files (optional, default=None): A list of files for which should be ignored and will not be in the output directory.
            no_expand_files (optional, default=None): A list of files which expansion should not be performed but which will be copied to the output directory.
            err_if_exists (optional, default=True): If True, raise an error if the output directory exists.

        Returns:
            Nothing.
        """
        source_dir = Path(source_dir).resolve()
        target_dir = Path(target_dir).resolve() if target_dir else Path.cwd()
        if is_debug('EXPANDER'):
            print('Using source directory:', source_dir)
            print('Using target directory:', target_dir)
        target_dir.mkdir(parents=True, exist_ok=not err_if_exists)
        for (root, _unused_dirs, files) in walk(source_dir):
            for file_name in files:
                if is_debug('EXPANDER'):
                    print('Checking', file_name, 'against', ignore_files)
                if file_name in ignore_files:
                    continue
                source_file = Path(root, file_name)
                target_file = target_dir / root.replace(str(source_dir), '').lstrip('\\' if WIN32 else '/') / file_name
                if file_name in no_expand_files:
                    if is_debug('EXPANDER'):
                        print(f'Copying {source_file} to {target_file} (root={root})')
                    copyfile(source_file, target_file)
                else:
                    if is_debug('EXPANDER'):
                        print(f'Expanding {source_file} to {target_file} ({root=})')
                    self.expand_file(source_file, target_file)

    def expand_file(self, in_file: PathName, out_file: PathName, /) -> None:
        """Perform an expansion on an entire file.

        Args:
            in_file: The name of the file on which to perform the expansions.
            out_file: The name of the output file for the expansion results.

        Returns:
            Nothing.
        """
        with open(in_file, encoding=DEFAULT_ENCODING) as in_stream:
            Path(out_file).parent.mkdir(parents=True, exist_ok=True)
            with open(out_file, 'w', encoding=DEFAULT_ENCODING) as out_stream:
                for line in in_stream:
                    line = self.expand(line)
                    out_stream.write(line)


def file_expander(in_file: PathName, out_file: PathName, /, *, var_dict: Optional[Dict[str, str]] = None, var_props: Any = None) -> None:
    """Quick function for one-time file expansion.

    Args:
        in_file: The input file.
        out_file: The output file.
        var_dict (optional, default=None): If not None, provides a dictionary of expansion values.
        var_props (optional, default=None): If not None, provides an object with properties to be used as expansion values.

    Returns:
        Nothing.
    """
    Expander(var_dict=var_dict, var_props=var_props).expand_file(in_file, out_file)


class Procedure:
    """Class to create a universal abstract interface for a procedure.

    Attributes:
        _SCHEMA_ATTR: The XML tag which contains the schema version.
        _REQUIRED_PROCEDURE_SCHEMA: The schema supported by this class.

        _HEADER_TAG: The XML tag which indicates the procedure header.
        _FLAGS_TAG: The XML tag which indicates the procedure flags.
        _DIRECTORIES_TAG: The XML tag which indicates the procedure directories.
        _ENVIRONMENTS_TAG: The XML tag which indicates the procedure environments.
        _STEPS_TAG: The XML tag which indicates the procedure steps.
        _LIBRARY_TAG: The XML tag which indicates the procedure step library.

        _COMMON_ENVIRONMENT: The XML tag which indicates the procedure common environment.
        _ENVIRONMENT_VARIABLE: The XML tag which indicates an environment variable.
    """
    _COMMON_ENVIRONMENT = 'common'
    _DIRECTORIES_TAG = 'directories'
    _ENVIRONMENTS_TAG = 'environments'
    _ENVIRONMENT_VARIABLE = 'Environment'
    _FLAGS_TAG = 'flags'
    _HEADER_TAG = 'header'
    _LIBRARY_TAG = 'step-library'
    _REQUIRED_PROCEDURE_SCHEMA = 1
    _SCHEMA_ATTR = 'schema'
    _STEPS_TAG = 'steps'

    def __init__(self, proc_file: PathName, /, *, output_format: OutputFormat = OutputFormat.html, variable_overrides: Optional[Dict[str, str]] = None):  # pylint: disable=too-many-locals
        """
        Args:
            proc_file: The procedure file.
            output_format: The output format.
            variable_overrides: A dictionary of values used to override values defined in the procedure file.

        Attributes:
            directories: The list of directories defined in the procedure file.
            environments: The dictionary of directories defined in the procedure file.
            expander: The expander object used to expand the procedure file.
            formatter: The formatter object used to format the output.
            header: The header of the output.
            library: The list of step libraries defined in the procedure file.
            output_format: The value of the output_format argument.
            steps: The list of steps defined in the procedure file.

        Raises:
            ProcedureError.BAD_SCHEMA: If the schema of the procedure file is not supported.
        """
        self.output_format = output_format
        self.formatter: Formatter
        self.expander: Expander

        if (schema := str_to_pythonval((xml_root := xml_parse(slurp(proc_file))).get(self._SCHEMA_ATTR, '0'))) != self._REQUIRED_PROCEDURE_SCHEMA:
            raise ProcedureError(ProcedureError.BAD_SCHEMA, schema=schema, expected=self._REQUIRED_PROCEDURE_SCHEMA)
        self.header = str(xml_root.findtext(self._HEADER_TAG)) if xml_root.findtext(self._HEADER_TAG) else ''
        flags = {f.tag: parse_flag(str(f.text)) for f in list(flags_element)} if (flags_element := xml_root.find(self._FLAGS_TAG)) else {}  # pylint: disable=used-before-assignment
        self.directories = [str(d.text) for d in list(directories_element)] if (directories_element := xml_root.find(self._DIRECTORIES_TAG)) else []  # pylint: disable=used-before-assignment
        self.steps = [Step(s) for s in list(steps_element)] if (steps_element := xml_root.find(self._STEPS_TAG)) else []  # pylint: disable=used-before-assignment
        self.library = {r.attrib[Step.NAME_ATTR]: Step(r) for r in list(library_element)} if (library_element := xml_root.find(self._LIBRARY_TAG)) else {}  # pylint: disable=used-before-assignment

        environments_element = xml_root.find(self._ENVIRONMENTS_TAG)
        self.environments: Dict = {e.tag: {v.tag: (v.text if v.text else '') for v in list(e)} for e in list(environments_element)} if environments_element else {}

        common_environment: Dict
        if self._COMMON_ENVIRONMENT in self.environments:
            common_environment = self.environments[self._COMMON_ENVIRONMENT]
            del self.environments[self._COMMON_ENVIRONMENT]
        else:
            common_environment = {}

        # Convert the flags to True/False variables in the common environment
        for (flag_name, flag_value) in flags.items():
            common_environment[flag_name] = flag_value

        # Create a default value of False for IsEnvironment of every environment
        for env in self.environments:
            common_environment['Is' + env] = False

        # Update the environments with the common environment values and
        # set the Environment variable and
        # set IsEnvironment to True for that environment
        for env in self.environments:
            env_dict = common_environment | self.environments[env]
            if variable_overrides:
                env_dict |= variable_overrides
            self.environments[env] = env_dict
            if self._ENVIRONMENT_VARIABLE not in self.environments[env]:
                self.environments[env][self._ENVIRONMENT_VARIABLE] = env
            self.environments[env]['Is' + env] = True

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def dump(self) -> Dict[str, str | Sequence | Dict]:
        """Dump out the procedure contents.

        Returns:
            The contents of the procedure as an ordered dictionary.
        """
        result: Dict = odict()
        result[self._HEADER_TAG] = self.header
        result[self._DIRECTORIES_TAG] = self.directories
        result[self._ENVIRONMENTS_TAG] = {e: v for (e, v) in self.environments.items()}  # pylint: disable=unnecessary-comprehension
        result[self._LIBRARY_TAG] = {r: s.dump() for (r, s) in self.library.items()}
        result[self._STEPS_TAG] = [s.dump() for s in self.steps]
        return result

    def expand(self, text: str, /) -> str:
        """Expand the Procedure.

        Args:
            text: The text of the procedure.

        Returns:
            The expanded procedure.

        Raises:
            ProcedureError.EXPANSION_ERROR: If there is an missing value expansion error.
        """
        try:
            return self.expander.expand(text)
        except ExpanderError as err:
            if err.code == ExpanderError.NO_REPLACEMENT.code:
                raise ProcedureError(ProcedureError.EXPANSION_ERROR, err=str(err), text=text) from err
            raise

    def expand_directories(self, env: str, /, destination_root: PathName, *, source_root: PathName = Path(), err_if_exists: bool = True) -> None:
        """Perform variable expansion on the directories defined in the procedure.

        Args:
            env: The environment for which to expand the directories.
            destination_root: That destination directory for the expansion.
            source_root (optional, default=None): Defined for recursion.
            err_if_exists (optional, default=True): If True, raise an error if destination_root exists.

        Returns:
            Nothing.
        """
        self.setup_expander(env)
        for dirname in self.directories:
            if not (dirpath := Path(dirname)).is_absolute():
                dirpath = Path(source_root) / dirpath
            self.expander.expand_directory(dirpath, Path(destination_root, dirname), err_if_exists=err_if_exists)

    def format(self, text: str, /) -> str:
        """Format an output line including hyperlinks.

        Args:
            text: The line of text to format.

        Returns:
            The formatted output line.
        """
        return self.formatter.format_hyperlinks(self.expand(text))

    def realize(self, env: str, /) -> str:
        """Realize the procedure for the specified environments based on the variables.

        Args:
            env: The environment for which to realize the procedure.

        Returns:
            The realized procedure.

        Raises:
            ProcedureError.BAD_FORMAT: If the format type is not defined.
        """
        header = footer = ''
        self.formatter = Formatter(self.output_format)
        self.setup_expander(env)
        match self.output_format:
            case OutputFormat.csv:
                header = f',{self.header.strip()} for {{var:Environment}}\n'
                footer = ''
            case OutputFormat.text:
                header = f'{self.header.strip()} for {{var:Environment}}\n'
                footer = ''
            case OutputFormat.html:
                header = '<html><meta http-equiv="Content-Type" content="text/html;charset=utf-8"><body><center>'
                header += f'<h1>{self.header} for {{var:Environment}}</h1></center><ol type="A">'
                footer = '</ol></body></html>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.output_format)

        content = ''
        for step in self.steps:
            text = self.realize_step(step)
            if text:
                content += self.formatter.eol + text
                self.formatter.increment()
        return self.format(header) + content + footer

    def realize_step(self, step: 'Step', /) -> str:  # pylint: disable=too-many-branches
        """Realize a step in the procedure.

        Args:
            step: The step to realize.

        Returns:
            The realized step.

        Raises:
            ProcedureError.BAD_LIBRARY: If the step specifies a library that is not defined.
        """
        if not self.expander.evaluate_expression(step.condition):
            return ''

        expander_vars_keeper = None
        if step.vars:
            expander_vars_keeper = deepcopy(self.expander.var_dict)
            self.expander.var_dict |= step.vars

        if step.lib_import:
            if step.lib_import not in self.library:
                raise ProcedureError(ProcedureError.BAD_LIBRARY, lib=step.lib_import)
            lib_step = deepcopy(self.library[step.lib_import])
            lib_step.repeat = step.repeat
            lib_step.vars |= step.vars
            if step.text:
                lib_step.text = step.text
            elif not lib_step.text:
                lib_step.text = lib_step.name
            step = lib_step

        output = ''
        if step.repeat:
            (variable, values_as_str) = step.repeat.split('=')
            values = [v.strip() for v in self.expand(values_as_str).split(',')]
            step_copy = deepcopy(step)
            step_copy.repeat = ''
            for value in values:
                step_copy.vars[variable] = value
                output += self.realize_step(step_copy)
        elif step.sub_steps:
            start_of_step = self.formatter.bol + self.format(step.text) + self.formatter.eol + self.formatter.bos
            self.formatter.indent()
            for sub_step in step.sub_steps:
                output += self.realize_step(sub_step)
            if output:
                output = start_of_step + output + self.formatter.eos
            self.formatter.outdent()
        else:
            if output := self.format(step.text):
                output = self.formatter.bol + output + self.formatter.eol
                self.formatter.increment()

        if expander_vars_keeper:
            self.expander.var_dict = expander_vars_keeper
        return output

    def setup_expander(self, environment: str, /) -> None:
        """Setup the Expander for the requested environment.

        Args:
            environment: The environment for which to setup the expander.

        Returns:
            Nothing.

        Raises:
            ProcedureError.BAD_ENVIRONMENT: If the requested environment is not defined.
        """
        if environment not in self.environments:
            raise ProcedureError(ProcedureError.BAD_ENVIRONMENT, env=environment)
        self.expander = Expander(var_dict=self.environments[environment])


class Step:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a procedure step.

    Attributes:
        NAME_ATTR: The XML attribute to indicate the step name.
        _CONDITION_ATTR: The XML attribute to indicate the step condition.
        _IMPORT_ATTR: The XML attribute to indicate the step import.
        _REPEAT_ATTR: The XML attribute to indicate the step repeat condition.
        _VARS_ATTR: The XML attribute to indicate the step variables.
    """
    _CONDITION_ATTR = 'condition'
    _IMPORT_ATTR = 'import'
    _REPEAT_ATTR = 'repeat'
    _VARS_ATTR = 'vars'

    NAME_ATTR = 'name'

    def __init__(self, step_def: Element, /):
        """
        Args:
            step_def: The dictionary which defined the step.

        Attributes:
            condition: The condition the determines if the step is emitted.
            lib_import: The library that is imported for the step.
            name: The step name.
            repeat: The repeat parameters for the step.
            sub_steps: The list of sub_steps for the step.
            text: The text emitted for the step.
            vars: The dictionary of variables defined for the step.
        """
        self.condition = step_def.get(self._CONDITION_ATTR, '')
        self.lib_import = step_def.get(self._IMPORT_ATTR, '')
        self.name = step_def.get(self.NAME_ATTR, '')
        self.repeat = step_def.get(self._REPEAT_ATTR, '')
        self.text = step_def.text.strip() if step_def.text else ''
        self.sub_steps = [Step(s) for s in list(step_def)]
        self.vars = {v.split('=')[0].strip(): v.split('=')[1].strip() for v in var.split(',')} if (var := step_def.get(self._VARS_ATTR, '')) else {}  # pylint: disable=used-before-assignment

    def dump(self) -> List:
        """Dump out the step contents.

        Returns:
            The contents of the step as an list.
        """
        return [f'{self.text}: import={self.lib_import}: condition={self.condition}: repeat={self.repeat}: vars={self.vars}'] + [str(s.dump()) for s in self.sub_steps]


def parse_flag(flag: str, /) -> Any:
    """Evaluate a parsing flag.

    Args:
        flag: The flag to evaluate.

    Return:
        The evaluated value for the flag.

    Raises:
        ProcedureError.BAD_FLAG: If the evaluated flag is not of type bool.
    """
    if not isinstance(value := str_to_pythonval(flag.lower().replace('0', 'False').replace('1', 'True').replace('yes', 'True').replace('no', 'False')), bool):
        raise ProcedureError(ProcedureError.BAD_FLAG, value=flag)
    return value

# cSpell:ignore odict fileutil pythonval postlim



"""This module provides utilities for creating command line menus."""

# Import standard modules
from dataclasses import dataclass
from typing import cast, List


@dataclass(frozen=True)
class MenuItem:
    """Class to represent a single menu item.

        Attributes:
            key: The key input for the menu selection.
            desc: The text description for the menu selection.
    """
    key: str
    desc: str


@dataclass
class Menu:
    """Class to create a universal abstract interface for a command-line menu.

        Attributes:
            items: The list of items for the menu.
            title (optional, default=_DEFAULT_TITLE): The title for the menu.
            prompt (optional, default=_DEFAULT_PROMPT): The prompt for the menu.
            invalid_msg (optional, default=_DEFAULT_INVALID_MESSAGE): The invalid choice message.
            multiselect (optional, default=False): If True, multiple options can be selected.
            ignorecase (optional, default=True): If True, menu input will be case insensitive.
    """
    items: List[MenuItem]
    title: str = '\nSelect one of the following\n'
    prompt: str = '-> '
    invalid_msg: str = '\nInvalid choice\n'
    multiselect: bool = False
    ignorecase: bool = True

    def show(self) -> str | List[str]:
        """Show the menu.

        Returns:
            The choice selected from the menu.
        """
        invalid_choice = True
        valid_choices = []
        menu = self.title + '\n'
        for item in self.items:
            valid_choices.append(item.key.upper() if self.ignorecase else item.key)
            menu += f'\t{item.key}. {item.desc}\n'

        choices = []
        while invalid_choice:
            print(menu)

            choices = [c.strip() for c in input(self.prompt).split(',')]
            if self.ignorecase:
                choices = [c.upper() for c in choices]

            invalid_choice = False
            for choice in choices:
                if choice not in valid_choices:
                    invalid_choice = True
            if not choices:
                invalid_choice = True
            if invalid_choice:
                print(self.invalid_msg)
        print()
        return choices if self.multiselect else choices[0]


@dataclass
class SimpleMenu(Menu):
    """A simplified version of the Menu class.

        Attributes:
            return_text (optional, default=False): If True the text of the menu selection will be returned instead of the key.
    """
    return_text: bool = False

    def __post_init__(self):
        self.items = [MenuItem(str(i), self.items[i - 1]) for i in range(1, len(self.items) + 1)] + [MenuItem('0', 'Exit')]

    def show(self) -> str:
        """Show the menu.

        Returns:
            The choice selected from the menu.
        """
        choice = cast(str, super().show())
        return self.items[int(choice) - 1].desc if self.return_text else choice
